// This file was generated by $generator from $template

#pragma once

#include "$prefixObject.h"
#include "$prefixObjects.h"

$begin_namespaces

template <>
class Object<Device> : public DefaultObject<Device>
{
public:
  Object(ANARIDevice d, ANARIObject handle);
  int getProperty(const char *propname,
    ANARIDataType type,
    void *mem,
    uint64_t size,
    ANARIWaitMask mask) override;

};

$end_namespaces
